from collections import defaultdict

from rest_framework import serializers

from arkindex.documents.models import Corpus, Element, ElementType
from arkindex.ponos.models import Farm
from arkindex.process.models import Process, ProcessMode
from arkindex.process.utils import get_default_farm


class BucketSerializer(serializers.Serializer):

    name = serializers.CharField()


class S3ImportSerializer(serializers.ModelSerializer):

    corpus_id = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        source="corpus",
        write_only=True,
        help_text="ID of the corpus to import S3 objects into. The user must have admin access on the corpus.",
        style={"base_template": "input.html"},
    )
    element_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        allow_null=True,
        required=False,
        write_only=True,
        source="element",
        help_text="Optional ID of a parent element to import into.",
        style={"base_template": "input.html"},
    )
    # We do not use SlugRelatedField here because their queryset depends on the `corpus_id` field;
    # we will perform the validation ourselves.
    folder_type = serializers.SlugField(
        help_text="Slug of the element type used for folders.",
        default="folder",
        write_only=True,
    )
    element_type = serializers.SlugField(
        help_text="Slug of the element type used for images.",
        default="page",
        write_only=True,
    )
    farm_id = serializers.PrimaryKeyRelatedField(
        queryset=Farm.objects.all(),
        write_only=True,
        allow_null=True,
        default=None,
        source="farm",
    )
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())
    mode = serializers.HiddenField(default=ProcessMode.S3)

    class Meta:
        model = Process
        fields = (
            "id",
            "bucket_name",
            "prefix",
            "corpus_id",
            "element_id",
            "folder_type",
            "element_type",
            "farm_id",
            "creator",
            "mode",
        )
        extra_kwargs = {
            "bucket_name": {"required": True, "allow_blank": False, "allow_null": False, "write_only": True},
            "prefix": {"write_only": True},
        }
        read_only_fields = ("id", )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "request" not in self.context:
            return
        user = self.context["request"].user
        corpora = Corpus.objects.admin(user)
        self.fields["corpus_id"].queryset = corpora
        self.fields["element_id"].queryset = Element.objects.filter(corpus__in=corpora)

    def validate_farm_id(self, farm):
        if farm is None:
            farm = get_default_farm()

        if not farm.is_available(self.context["request"].user):
            raise serializers.ValidationError(["You do not have access to this farm."])

        return farm

    def validate(self, data):
        data = super().validate(data)
        errors = defaultdict(list)
        corpus = data["corpus"]
        element = data.get("element")

        if element is not None and element.corpus_id != corpus.id:
            errors["element"].append(f"This element is not part of the corpus {corpus.id}")

        try:
            data["element_type"] = corpus.types.get(slug=data["element_type"], folder=False)
        except ElementType.DoesNotExist:
            errors["element_type"].append(f"Corpus {corpus.id} does not have a {data['element_type']!r} type, or it is a folder type")

        try:
            data["folder_type"] = corpus.types.get(slug=data["folder_type"], folder=True)
        except ElementType.DoesNotExist:
            errors["folder_type"].append(f"Corpus {corpus.id} does not have a {data['folder_type']!r} type, or it is not a folder type")

        if errors:
            raise serializers.ValidationError(errors)

        return data

    def create(self, *args, **kwargs):
        instance = super().create(*args, **kwargs)
        instance.run()
        return instance
