import uuid
from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Artifact, Farm
from arkindex.ponos.models import State as PonosState
from arkindex.process.models import (
    FeatureUsage,
    Process,
    ProcessMode,
    WorkerConfiguration,
    WorkerRun,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureTestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role


class TestCreateTrainingProcess(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.train_folder = cls.corpus.elements.get(name="Volume 1")
        cls.validation_folder = cls.corpus.elements.get(name="Volume 2")
        cls.test_folder = cls.corpus.elements.create(type=cls.train_folder.type, name="Volume 3")
        cls.training_worker_version = WorkerVersion.objects.get(worker__slug="worker-gpu")
        cls.training_worker_version.worker.memberships.create(user=cls.user, level=Role.Admin.value)
        cls.training_worker_version.gpu_usage = FeatureUsage.Supported
        cls.training_worker_version.docker_image_iid = 1337
        cls.training_worker_version.docker_image = Artifact.objects.first()
        cls.training_worker_version.save()
        # Build a configuration and a model version as they does not exists in the fixture yet
        cls.configuration = WorkerConfiguration.objects.create(
            name="Test",
            configuration={},
            configuration_hash=42,
            worker=cls.training_worker_version.worker,
        )
        cls.model = Model.objects.create(
            name="Generic model",
            public=False,
        )
        cls.model.compatible_workers.add(cls.training_worker_version.worker)
        cls.model.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            hash="42",
            archive_hash="42",
            size=1337,
        )

        cls.default_farm = Farm.objects.create(name="Crypto farm")
        cls.default_farm.memberships.create(user=cls.user, level=Role.Guest.value)
        cls.other_farm = Farm.objects.get(name="Wheat farm")

    def setUp(self):
        super().setUp()
        # Base payload with the required fields
        self.base_payload = {
            "name": "Test training",
            "corpus_id": str(self.corpus.id),
            "train_folder_id": str(self.train_folder.id),
            "worker_version_id": str(self.training_worker_version.id),
            "model_id": str(self.model.id),
        }

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_required_parameters(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:process-training"), {})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "corpus_id": ["This field is required."],
            "name": ["This field is required."],
            "train_folder_id": ["This field is required."],
            "worker_version_id": ["This field is required."],
            "model_id": ["This field is required."],
        })

    def test_invalid_values(self):
        fields = [
            *self.base_payload.keys(),
            "validation_folder_id",
            "test_folder_id",
            "worker_version_id",
            "model_version_id",
            "worker_configuration_id",
            "use_gpu",
        ]
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:process-training"),
                # Cast every parameter as an integer
                {field: "42" for field in fields},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        expected_errors = {field: ["“42” is not a valid UUID."] for field in fields}
        # Model version ID use a serrializers.UUIDField
        expected_errors["model_version_id"] = ["Must be a valid UUID."]
        # Name can be any string and use_gpu is a boolean
        del expected_errors["name"]
        expected_errors["use_gpu"] = ["Must be a valid boolean."]
        self.assertEqual(response.json(), expected_errors)

    def test_corpus_right(self):
        """
        The user must be a contributor to the corpus
        """
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "corpus_id": [f'Invalid pk "{self.corpus.id}" - object does not exist.'],
            "train_folder_id": [f'Invalid pk "{self.train_folder.id}" - object does not exist.'],
        })

    def test_worker_version_right(self):
        """
        The user must be an admin or contributor of the training worker, as new trained models will be published
        """
        self.training_worker_version.worker.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "worker_version_id": [f'Invalid pk "{self.training_worker_version.id}" - object does not exist.'],
        })

    def test_model_right(self):
        self.model.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "model_id": [f'Invalid pk "{self.model.id}" - object does not exist.'],
        })

    def test_model_version_belongs_to_the_model(self):
        other_model_version = ModelVersion.objects.create(
            model=Model.objects.create(name="Secret model", public=False),
            state=ModelVersionState.Available,
            hash="111",
            archive_hash="222",
            size=333,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(15):
            response = self.client.post(reverse("api:process-training"), {
                **self.base_payload,
                "model_version_id": str(other_model_version.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "model_version_id": ["Object does not exist or is not a version of the specified model"]
        })

    def test_folders_validation(self):
        self.client.force_login(self.user)
        train_id, valid_id, test_id = uuid.uuid4(), uuid.uuid4(), uuid.uuid4()
        with self.assertNumQueries(16):
            response = self.client.post(reverse("api:process-training"), {
                **self.base_payload,
                "train_folder_id": train_id,
                "validation_folder_id": valid_id,
                "test_folder_id": test_id,
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "train_folder_id": [f'Invalid pk "{train_id}" - object does not exist.'],
            "validation_folder_id": [f'Invalid pk "{valid_id}" - object does not exist.'],
            "test_folder_id": [f'Invalid pk "{test_id}" - object does not exist.'],
        })

    def test_folders_similar(self):
        self.client.force_login(self.user)
        folder_id = self.base_payload["train_folder_id"]
        with self.assertNumQueries(16):
            response = self.client.post(reverse("api:process-training"), {
                **self.base_payload,
                "validation_folder_id": folder_id,
                "test_folder_id": folder_id,
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "__all__": ["Train, validation and test folders must be different."],
        })

    def test_worker_version_not_available(self):
        self.client.force_login(self.user)
        # Every state other than Available should cause this error
        states = set(WorkerVersionState) - {WorkerVersionState.Available}
        for state in states:
            with self.subTest(state=state):
                self.training_worker_version.state = state
                self.training_worker_version.save()

                with self.assertNumQueries(14):
                    response = self.client.post(reverse("api:process-training"), self.base_payload)
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "worker_version_id": ["This worker version is not available."]
                })

    def test_worker_version_gpu_required(self):
        """
        An error is raised in case the GPU mode is incompatible with the worker
        """
        self.client.force_login(self.user)
        self.training_worker_version.gpu_usage = FeatureUsage.Required
        self.training_worker_version.save()
        with self.assertNumQueries(14):
            response = self.client.post(reverse("api:process-training"), {
                **self.base_payload,
                "use_gpu": False,
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "__all__": ["This worker version does only support GPU mode."]
        })

    def test_worker_version_gpu_not_supported(self):
        """
        An error is raised in case the GPU mode is incompatible with the worker
        """
        self.client.force_login(self.user)
        self.training_worker_version.gpu_usage = FeatureUsage.Disabled
        self.training_worker_version.save()
        with self.assertNumQueries(14):
            response = self.client.post(reverse("api:process-training"), {
                **self.base_payload,
                "use_gpu": True,
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "__all__": ["This worker version does not support GPU mode."]
        })

    def test_worker_missing_docker_image(self):
        """
        The worker version must contain a docker image
        """
        self.client.force_login(self.user)
        self.training_worker_version.docker_image_iid = None
        self.training_worker_version.save()
        with self.assertNumQueries(14):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "worker_version_id": ["This version has no Docker image, which is required for the training."]
        })

    @override_settings(PONOS_DEFAULT_ENV={})
    @patch("arkindex.process.builder.task_token_default")
    def test_create_training(self, token_mock):
        token_mock.return_value = "12345"

        self.client.force_login(self.user)
        training_process = Process.objects.filter(name="Test training")
        self.assertEqual(training_process.count(), 0)

        with self.assertNumQueries(25):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(training_process.count(), 1)
        training_process = training_process.get()
        self.assertEqual(training_process.creator_id, self.user.id)
        self.assertEqual(training_process.farm, self.default_farm)

        self.assertEqual(response.json(), {
            "id": str(training_process.id),
            "mode": ProcessMode.Training.value,
            "activity_state": "disabled",
            "corpus": str(self.corpus.id),
            "element": None,
            "element_name_contains": None,
            "element_type": None,
            "files": [],
            "folder_type": None,
            "load_children": False,
            "ml_class_id": None,
            "model_id": str(self.model.id),
            "name": "Test training",
            "revision": None,
            "state": PonosState.Unscheduled.value,
            "template_id": None,
            "test_folder_id": None,
            "train_folder_id": str(self.train_folder.id),
            "use_cache": False,
            "use_gpu": False,
            "validation_folder_id": None,
        })

        worker_run = training_process.worker_runs.get()

        training_task = training_process.tasks.get()
        self.assertEqual(training_task.slug, "training")
        self.assertDictEqual(training_task.env, {
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_PROCESS_ID": str(training_process.id),
            "ARKINDEX_TASK_TOKEN": "12345",
            "ARKINDEX_WORKER_RUN_ID": str(worker_run.id),
        })
        self.assertEqual(training_task.image, "1337")
        self.assertEqual(training_task.command, None)
        self.assertEqual(training_task.image_artifact.id, self.training_worker_version.docker_image_id)
        self.assertEqual(training_task.shm_size, None)
        self.assertEqual(training_task.requires_gpu, False)

        # Check worker run properties
        self.assertEqual(worker_run.version_id, self.training_worker_version.id)
        self.assertEqual(worker_run.model_version_id, None)

    @override_settings(PONOS_DEFAULT_ENV={})
    def test_create_training_worker_contributor(self):
        self.training_worker_version.worker.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)
        training_process = Process.objects.filter(name="Test training")
        self.assertEqual(training_process.count(), 0)
        with self.assertNumQueries(24):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(training_process.count(), 1)

    @override_settings(PONOS_DEFAULT_ENV={})
    def test_create_training_worker_public(self):
        self.training_worker_version.worker.public = True
        self.training_worker_version.worker.save()
        self.training_worker_version.worker.memberships.filter(user=self.user).delete()
        self.client.force_login(self.user)
        training_process = Process.objects.filter(name="Test training")
        self.assertEqual(training_process.count(), 0)
        with self.assertNumQueries(24):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(training_process.count(), 1)

    @override_settings(PONOS_DEFAULT_ENV={})
    @patch("arkindex.process.models.task_token_default")
    def test_create_training_extra_parameters(self, token_mock):
        token_mock.return_value = "12345"
        self.client.force_login(self.user)
        training_process = Process.objects.filter(name="Test training")
        self.assertEqual(training_process.count(), 0)

        with self.assertNumQueries(29):
            response = self.client.post(reverse("api:process-training"), {
                **self.base_payload,
                "validation_folder_id": str(self.validation_folder.id),
                "test_folder_id": str(self.test_folder.id),
                "model_version_id": str(self.model_version.id),
                "worker_configuration_id": str(self.configuration.id),
                "use_gpu": True,
                "farm_id": str(self.other_farm.id),
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(training_process.count(), 1)
        training_process = training_process.get()
        self.assertEqual(response.json(), {
            "id": str(training_process.id),
            "mode": ProcessMode.Training.value,
            "activity_state": "disabled",
            "corpus": str(self.corpus.id),
            "element": None,
            "element_name_contains": None,
            "element_type": None,
            "files": [],
            "folder_type": None,
            "load_children": False,
            "ml_class_id": None,
            "model_id": str(self.model.id),
            "name": "Test training",
            "revision": None,
            "state": PonosState.Unscheduled.value,
            "template_id": None,
            "test_folder_id": str(self.test_folder.id),
            "train_folder_id": str(self.train_folder.id),
            "use_cache": False,
            "use_gpu": True,
            "validation_folder_id": str(self.validation_folder.id),
        })
        self.assertEqual(training_process.creator_id, self.user.id)
        self.assertEqual(training_process.farm, self.other_farm)
        task = training_process.tasks.get()
        self.assertEqual(sorted(task.env.keys()), [
            "ARKINDEX_CORPUS_ID",
            "ARKINDEX_PROCESS_ID",
            "ARKINDEX_TASK_TOKEN",
            "ARKINDEX_WORKER_RUN_ID",
        ])
        self.assertEqual(task.requires_gpu, True)
        self.assertEqual(task.env["ARKINDEX_TASK_TOKEN"], task.token)
        # Check worker run properties
        self.assertEqual(str(training_process.worker_runs.get().id), task.env["ARKINDEX_WORKER_RUN_ID"])
        worker_run = WorkerRun.objects.get(id=task.env["ARKINDEX_WORKER_RUN_ID"])
        self.assertEqual(worker_run.version_id, self.training_worker_version.id)
        self.assertEqual(worker_run.model_version, self.model_version)

    @override_settings(PONOS_DEFAULT_ENV={})
    @patch("arkindex.process.builder.task_token_default")
    def test_create_training_shm_size(self, token_mock):
        token_mock.return_value = "12345"

        self.client.force_login(self.user)
        self.training_worker_version.configuration = {"docker": {"shm_size": "999G"}}
        self.assertEqual(self.training_worker_version.docker_shm_size, "999G")
        self.training_worker_version.save()

        with self.assertNumQueries(25):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        training_process = Process.objects.get(mode=ProcessMode.Training, name="Test training")
        worker_run = training_process.worker_runs.get()

        task = training_process.tasks.get()
        self.assertEqual(task.env, {
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_PROCESS_ID": str(training_process.id),
            "ARKINDEX_TASK_TOKEN": "12345",
            "ARKINDEX_WORKER_RUN_ID": str(worker_run.id),
        })
        self.assertEqual(task.slug, "training")
        self.assertEqual(task.image, "1337")
        self.assertEqual(task.command, None)
        self.assertEqual(task.image_artifact.id, self.training_worker_version.docker_image_id)
        self.assertEqual(task.requires_gpu, False)
        self.assertEqual(task.shm_size, "999G")

    def test_default_farm_guest(self):
        self.default_farm.memberships.filter(user=self.user).delete()
        self.client.force_login(self.user)

        with self.assertNumQueries(13):
            response = self.client.post(reverse("api:process-training"), self.base_payload)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "farm_id": ["You do not have access to this farm."],
        })

        self.assertFalse(Process.objects.filter(name="Test training").exists())

    def test_farm_guest(self):
        self.other_farm.memberships.filter(user=self.user).delete()
        self.client.force_login(self.user)

        with self.assertNumQueries(13):
            response = self.client.post(reverse("api:process-training"), {
                **self.base_payload,
                "farm_id": str(self.other_farm.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "farm_id": ["You do not have access to this farm."],
        })

        self.assertFalse(Process.objects.filter(name="Test training").exists())
